#include "trios.h"
#include "haplotype.h"
#include "parameters.h"

void displayPhase2(CPhase & p);

void CBox::enumeratePhase(Window & win)
{

    if (par::debug)
	cerr << "\n enumeratePhase() \n";

    // populate hap with all consistent haplotype pairs, genotype data could be missing

    // clear all phases initially
    clearPhase();

    // For H heterozygous sites, there are 2^(H-1) possible haplotypes
    // After 1st heterozygous site, add all flips of subsequent heterozygous sites

    // list of heterozygous sites
    vector<int> het;

    // list of missing sites
    vector<int> miss;

    // Temporary Haplotype Pair
    int n = ind::nloci;
    // and check
    if (n != win.alt.size()) error("Internal error: mismatch # of loci");
    
    // Copy diplotype information into temporary haplotypes
    CHaplotype a(n), b(n), save_a(n), save_b(n);
    
    for (int i=1; i <= n; i++)
    {
	a.pos(i) = person->loc(win.alt[i-1]).one;
	b.pos(i) = person->loc(win.alt[i-1]).two;

	save_a.pos(i) = person->loc(win.alt[i-1]).one;
	save_b.pos(i) = person->loc(win.alt[i-1]).two;
    }

    // move up to 1st heterozygous site
    int h = 0;
    for (int p = 1 ; p <= n ; p++) 
	if ( a.pos(p) != b.pos(p) )
	   { h = p; break; } 
    
    // add het sites
    for (int p = h+1 ; p <= n ; p++)
	if ( a.pos(p) != b.pos(p) )
	    het.push_back(p);

    // add all missing sites
    for (int p = 1 ; p <= n ; p++)
	if ( a.pos(p) == "0" ) 
	    miss.push_back(p);

    int nhap = int(pow(2.0,int(het.size())));
    int nmiss = int(pow(2.0,int(2*miss.size()))); 

        // if no missing (2^0 = 1) 
    if (miss.size() == 0) nmiss = 0;

    // append flipped heterozygous sites 
    for (int i = 0 ; i < nhap ; i++)
    {
	int gc = graycode(i,het.size());
	
	a = save_a;
	b = save_b;
	
	for (int p = 0; p < het.size(); p++)
	    if (gc & int(pow(2.0,p))) 
	    {
		a.pos(het[p]) = save_b.pos(het[p]);
		b.pos(het[p]) = save_a.pos(het[p]);
	    }
	
	// If no missing genotypes just add these straight
	if (nmiss == 0)
	{
	    addPhase(a,b);
	    addPhase(b,a);
	}
	else
	{
	    // Otherwsie append all possible missing sites
	    for (int i = 0 ; i < nmiss ; i++)    
	    {
		int gc = graycode(i,nmiss);
		
		a = save_a;
		b = save_b;
		
		for (int p = 0; p < miss.size() ; p++)
		{

		    // missing genotype, first position
		    if (gc & int(pow(2.0,p*2)))
			a.pos(miss[p]) = 1;
		    else
			a.pos(miss[p]) = 2;

		    if (gc & int(pow(2.0,p*2+1)))
			b.pos(miss[p]) = 1;
		    else
			b.pos(miss[p]) = 2;
		}
		
		addPhase(a,b);
		addPhase(b,a);
		
	    }
	    
	}
	
    }

}


void CBox::addPhase(CHaplotype pat, CHaplotype mat, double prob)
{
    CPhase new_phase(pat,mat,prob);
    P.push_back(new_phase);
    nphase++;
}

void CBox::removePhase(int n)
{
    P.erase(P.begin()+n-1);
    nphase--;
}

void CBox::clearPhase()
{
    P.clear();
    nphase=0;
}



void phaseChildGivenParents(Driver &d)
{

    if (par::debug)
	cerr << "\n phaseChildGivenParents() \n";

    ////////////////////////////////////////////////
    // Consider each family
    for (int f= 1 ; f <= d.sample.countFamily(); f++)
    {
	
	// Skip if this is an UNRELATED
	if ( d.sample.fam(f)->nInFamily() == 1) continue;

	// ASSUMPTION: otherwise, assume this family is a trio

	// find the child
	CIndividual & kid = d.sample.fam(f)->getChildFromTrio();
	
	// and enumerate all possible phases in P (need to pass Window
	// specification also, so that we can get child genotypes from
	// CSample

	// Calc conditional on on Parental GENO instead...

	kid.box->PP.clear();

	// Pointers to father and mother
	// Father and mother phases will have been generated by E-M
	CIndividual & father = *kid.pat;
	CIndividual & mother = *kid.mat;


	//////////////////////////////////////////////
	// Calculate P(H_C | G_C, H_F, H_M )

	// consider each pairwise combination of parental phases
	for (int j=1; j <= father.box->nphase; j++)
	    for (int k=1; k <= mother.box->nphase; k++)
	    {
		if (father.box->phase(j).prob == 0 || 
		    mother.box->phase(k).prob == 0)
		{
		    if (par::dump)
			cout << "\n skipping impossible config \n";

		    continue;
		}
		
		if (par::dump)
		{
		    cout << "\n fam " << f << "\n";
		    cout << "pat = "; displayPhase2(father.box->phase(j));
		    cout << "\nmat = ";
		    displayPhase2(mother.box->phase(k));
		}
			     
                // Reset probs (these are conditional on parental haps)
		double prob = 0;	
		
		// Paternal phase
		CPhase & fphase = father.box->phase(j);
		
		// Maternal phase
		CPhase & mphase = mother.box->phase(k);

		// start of CHANGE -- see 'CHANGE' above ... 
		// we now enumerate conditional on parental phase
		if (par::debug)
		    cerr << "   about to enumerateGivenParents() for trio " << father.family->getID() << "\n";
    
		kid.box->P = enumerateGivenParents(fphase,mphase,d.locmap.theta(),par::recomb);
		// remove phases inconsistent with offspring genotype
		// i.e. Do not add if no possibility of child given child
		int p0=0;
		while (p0 < kid.box->P.size())
		{
		    bool consistent = true;
		    int lp = 1;
		    for (int l=1; l <= ind::nloci; l++)
		    {
			string g1, g2, p1, p2;
			if (!par::usat) 
			{
			    g1 = kid.loc(d.win[Window::n]->alt[l-1]).one;
			    g2 = kid.loc(d.win[Window::n]->alt[l-1]).two;
			    p1 = kid.box->P[p0].paternal.pos(l);
			    p2 = kid.box->P[p0].maternal.pos(l);

			    if ( ( g1+g2 != p1+p2 && g2+g1 != p1+p2  ) && g1+g2 != "00" ) 
				consistent = false;
			}
			else // uSat mode on
			{
			    int nl = d.win[Window::n]->alt[l-1]-1;
			    for (int l2=0; l2<d.acount[nl]; l2++)
			    {
				g1 = kid.loc(d.downcode[nl][l2]).one;
				g2 = kid.loc(d.downcode[nl][l2]).two;
				p1 = kid.box->P[p0].paternal.pos(lp);
				p2 = kid.box->P[p0].maternal.pos(lp);
				lp++;

				if ( ( g1+g2 != p1+p2 && g2+g1 != p1+p2  ) && g1+g2 != "00" ) 
				    consistent = false;
			    }
			}
			

			if (par::dump)
			    cout << " g1="<<g1
				 << " g2="<<g2
				 << " p1="<<p1
				 << " p2="<<p2
				 << " con="<<consistent<<"\n";

		    }
		    
		if (!consistent)
			kid.box->P.erase(kid.box->P.begin()+p0);
		    else p0++;
		}
		
		// count up
		kid.box->nphase = kid.box->P.size();
		// end of CHANGE

		// consider each kid phase
		for (int i = 1 ; i <= kid.box->nPhase() ; i++ )
		{
		    CPhase & phase = kid.box->phase(i);
		    phase.prob = phase.probGivenParents(fphase,
							mphase,
							d.locmap.theta(),
							par::recomb);
		    prob += phase.prob;
		    
		    if (par::dump)
		    {
			cout << "\n potential offspring = ";
			displayPhase2(phase);
		    }

		}
		
		// Do not add if no possibility of child given parents
		if (prob == 0) continue;
		
		// Make phase probabilities conditional 
		for (int i = 1 ; i <= kid.box->nPhase() ; i++ )
		{
		    CPhase & phase = kid.box->phase(i);
		    phase.prob /= prob;
		}
		
		// Add this to list
				
		CParentalPhase new_phase;
//		new_phase.prob = prob;     // no need to duplicate... revise this l8r
		new_phase.father = fphase;
		new_phase.mother = mphase;
		new_phase.P = kid.box->phase();
		
		// remove incosistent phases before adding, 
		// CHECK ;; see if this hasn't already been done...
 		int p=0;
 		while (p < new_phase.P.size())
 		{
 		    if (new_phase.P[p].prob == 0) 
 			new_phase.P.erase(new_phase.P.begin()+p);
 		    else p++;
 		}
		
		// add to list
		kid.box->PP.push_back(new_phase);
		
	    }
	
    }
}



void displayPhase2(CPhase & p)
 {
     int l = p.paternal.pos().size();
     for (int j=1; j<=l; j++)
 	cout << p.paternal.pos(j);
     cout << "\\";
     for (int j=1; j<=l; j++)
 	cout << p.maternal.pos(j);
     cout << "\t" << p.prob;     
     if (p.prob==0) cout << " * ";
 }



vector<CPhase> enumerateGivenParents(CPhase & dad, CPhase & mum, CMatrix & theta, int mr)
{



    vector<CPhase> all_phase;

    // get length 
    int n = dad.paternal.n; 
    int niv = int(pow(2.0,n));
    vector< vector<bool> > iv;

    if (mr == 0)
    {
	niv = 2;
	vector<bool> t(n);
	
	for (int i=0; i < n; i++)
	    t[i] = false;
	iv.push_back(t);
	
	for (int i=0; i < n; i++)
	    t[i] = true;
	iv.push_back(t);
	
    }
    else
    {
	// construct all 2^n possible inheritance vectors
	        
	for (int i = 0 ; i < niv ; i++)
	{
	    int gc = graycode(i,n);
	    
	    vector<bool> t(n);
	    iv.push_back(t);
	    
	    // each position
	    for (int p = 0; p < n; p++)
		if (gc & int(pow(2.0,p)))
		    iv[i][p] = true;
		else 
		    iv[i][p] = false;		
	}
	
    }

    // consider each IV 
    for (int iv_pat = 0 ; iv_pat < niv ; iv_pat++)
	for (int iv_mat = 0 ; iv_mat < niv ; iv_mat++)
	{
	    
	    // Is this IV consistent with max allowed recombinations?
	    int pat_r = 0;
	    int mat_r = 0;
	    for (int p=1; p<n; p++)
	    {
		if (iv[iv_pat][p] != iv[iv_pat][p-1]) pat_r++;
		if (iv[iv_mat][p] != iv[iv_mat][p-1]) mat_r++;
	    }
	    // skip if exceeds recombination limit
	    if (pat_r > mr || mat_r > mr) continue;
	    
	    
	    // Is this IV consistent with observed offspring haplotype?
	    
	    CPhase implied(n);
	    
	    for (int p = 1 ; p <= n ; p++)
	    {
		// implied offspring genotype
		
		if (iv[iv_pat][p-1])
		    implied.paternal.pos(p) = dad.paternal.pos(p);
		else
		    implied.paternal.pos(p) = dad.maternal.pos(p);
		
		
		if (iv[iv_mat][p-1])
		    implied.maternal.pos(p) = mum.paternal.pos(p);
		else
		    implied.maternal.pos(p) = mum.maternal.pos(p);
	    }
	    
	    
	    
	    // Calculate probability of child haplotype given parental haplotypes
	    
	    double prob_pat = 0.5;
	    double prob_mat = 0.5;
	    
	    for (int p = 1 ; p < n ; p++)
	    {
		// No transition
		if (iv[iv_pat][p-1] == iv[iv_pat][(p+1)-1]) 
		    prob_pat *= 1 - theta[p][p+1];
		else // Transition
		    prob_pat *= theta[p][p+1];

		// No transition
		if (iv[iv_mat][p-1] == iv[iv_mat][(p+1)-1]) 
		    prob_mat *= 1 - theta[p][p+1];
		else // Transition
		    prob_mat *= theta[p][p+1];

	    }

	    implied.prob = prob_pat * prob_mat;
	    
	    all_phase.push_back(implied);

	}

    return all_phase;

}





//////////////////////
//  Example of recombination


//       HAPLOTYPES.....    GENOTYPES
//  PAT  112211 / 111222    1/1 1/1 1/2 2/2 1/2 1/2
//  MAT  222112 / 112211    1/2 1/2 2/2 1/2 1/1 1/2

//  KID                     1/1 1/1 1/2 2/2 1/2 2/2   
//        PAT 111222
//        mat 112212

// PAT     1-1-1-2-2-2   
// MAT     1-1-2-2/1-2    4:2 for 112211 : 222112

// so split this phase into 
